/* -----  RSTRlintHeap.h  ----- */
#//
#// Copyright (c) 2017 - 2018 Norbert Beckmann  <nbinf@kabelmail.de>
#//
#// Licensed under the Apache License, Version 2.0 (see root directory)
#//
#ifndef __RSTRlintHeap_h
#define __RSTRlintHeap_h


/**:   Binary Heap of Rlint
       ====================                                             **/


/**    Implementation:  Norbert Beckmann
              Version:  1.1
                 Date:  05/18                                           **/

/** Main Principle:
    ===============
    A binary heap constitutes a priority queue, the first element either
    being the minimum one (Min-Heap) or being the maximum one (Max-Heap).
    Binary heaps are usually implemented in an array.
    The cost for inserting a new element and removing the first element
    is O(log n), n: number of elements in the heap.
    
    Level: bottom.
                                                                        **/

#include "RSTStdTypes.h"


/* ----- types ----- */

typedef t_RlintHeap  *t_liH;

        /* abstract data type t_liH,
           heap of Rlint. */


/* ----- functions ----- */

/*** ---------- Functions applicable to heaps generally ---------- */

boolean LIH_New(t_liH *hp,
                Rint max);

/* demands the address of a variable of type t_liH,
   initializes the heap hp and attempts to allocate space for max elements,
   returns TRUE if hp is ready for use,
   returns FALSE otherwise */


void LIH_Dispose(t_liH *hp);

/* deallocates the memory of hp, then sets hp to NULL */


void LIH_Empty(t_liH hp);

/* empties hp, keeps its memory and capacity */


Rint LIH_Qty(t_liH hp);

/* returns the number of elements of hp */


Rint LIH_MaxQty(t_liH hp);

/* returns the maximum number of elements of hp */


/* --------------------------- General Note ------------------------------- */
/* A heap (re-)constitutes as a Min-Heap or Max-Heap, depending on whether  */
/* minLIH_Insert or maxLIH_Insert is used to insert the second element      */
/* into it.                                                                 */
/*                                                                          */
/* Functions for Min-Heaps must not be applied to Max-Heaps and vice versa. */
/*                                                                          */
/* This is a lower level module which should be fast and simple. As a       */
/* consequence, violation of the above rule is neither reported nor even    */
/* detected, though damaging consistency.                                   */


/*** ------------------ Functions for Min-Heaps ------------------ */

boolean minLIH_Insert(t_liH minHp, Rlint elem);

/* attempts to insert elem into minHp ordered by INCREASING values,
   building up a Min-Heap,
   returns TRUE if elem was inserted (capacity was available),
   returns FALSE otherwise */


boolean minLIH_Pop(t_liH minHp, Rlint *elem);

/* attempts to remove the first, i.e. MINIMAL element from minHp, a Min-Heap,
   returns TRUE and provides the element in elem if minHp is not empty,
   returns FALSE otherwise */


/*** ------------------ Functions for Max-Heaps ------------------ */

boolean maxLIH_Insert(t_liH maxHp, Rlint elem);

/* attempts to insert elem into maxHp ordered by DECREASING values,
   building up a Max-Heap,
   returns TRUE if elem was inserted (capacity was available),
   returns FALSE otherwise */


boolean maxLIH_Pop(t_liH maxHp, Rlint *elem);

/* attempts to remove the first, i.e. MAXIMAL element from maxHp, a Max-Heap,
   returns TRUE and provides the element in elem if maxHp is not empty,
   returns FALSE otherwise */


#endif /* __RSTRlintHeap_h */

